@@ -80,7 +80,7 @@ XSPROTO(CAIXS_entersub_wrapper) {
 
 #ifdef OPTIMIZE_OPMETHOD
 
-template <AccessorTypes type> static
+template <AccessorTypes type, int optype> static
 OP *
 CAIXS_opmethod_wrapper(pTHX) {
     dSP;
@@ -88,6 +88,11 @@ CAIXS_opmethod_wrapper(pTHX) {
     SV* self = PL_stack_base + TOPMARK == SP ? (SV*)NULL : *(PL_stack_base + TOPMARK + 1);
     HV* stash = NULL;
 
+    /*
+        this block isn't required for the 'goto gotcv' case, but skipping it
+        (or swapping those blocks) will make unstealing inside 'goto gotcv' block impossible,
+        thus requiring additional check in the fast case, and subref is a failure anyway most of the times
+    */
 #ifndef GV_CACHE_ONLY
     if (LIKELY(self != NULL)) {
         SvGETMAGIC(self);
@@ -116,44 +121,69 @@ CAIXS_opmethod_wrapper(pTHX) {
         }
     }
 
-    /* SvTYPE comes from 5.22 */
-    if (UNLIKELY(!stash || SvTYPE(stash) != SVt_PVHV)) {
-        OP_UNSTEAL(OP_METHOD_NAMED);
-    }
-
+    SV* meth;
     CV* cv = NULL;
-    GV* gv;
-    SV* meth = cSVOPx_sv(PL_op);
+    U32 hash;
+
+    if (optype == OP_METHOD) {
+        meth = TOPs;
+        if (SvROK(meth)) {
+            SV* const rmeth = SvRV(meth);
+            if (SvTYPE(rmeth) == SVt_PVCV) {
+                cv = (CV*)rmeth;
+                goto gotcv; /* we don't care about the 'stash' var here */
+            }
+        }
+
+        hash = 0;
+
+    } else if (optype == OP_METHOD_NAMED) {
+        meth = cSVOPx_sv(PL_op);
 
 #ifndef GV_CACHE_ONLY
-    const U32 hash = SvSHARED_HASH(meth); /* OP_METHOD doesn't have hash and skips this, but only below 5.22 */
+        hash = SvSHARED_HASH(meth);
 #else
-    const U32 hash = 0;
+        hash = 0;
 #endif
-    HE* he = hv_fetch_ent(stash, meth, 0, hash);
-    if (he) {
-        gv = (GV*)(HeVAL(he));
+    }
+
+    /* SvTYPE check appeared only since 5.22, but execute it for all perls nevertheless */
+    if (UNLIKELY(!stash || SvTYPE(stash) != SVt_PVHV)) {
+        OP_UNSTEAL(optype);
+    }
+
+    HE* he; /* to allow 'goto' to jump over this */
+    if ((he = hv_fetch_ent(stash, meth, 0, hash))) {
+        GV* gv = (GV*)(HeVAL(he));
         if (isGV(gv) && GvCV(gv) && (!GvCVGEN(gv) || GvCVGEN(gv) == (PL_sub_generation + HvMROMETA(stash)->cache_gen))) {
             cv = GvCV(gv);
         }
     }
 
     if (UNLIKELY(!cv)) {
-        gv = gv_fetchmethod_sv_flags(stash, meth, GV_AUTOLOAD|GV_CROAK);
+        GV* gv = gv_fetchmethod_sv_flags(stash, meth, GV_AUTOLOAD|GV_CROAK);
         assert(gv);
 
         cv = isGV(gv) ? GvCV(gv) : (CV*)gv;
         assert(cv);
     }
 
+gotcv:
     if (LIKELY(CvXSUB(cv) == (XSUBADDR_t)&CAIXS_entersub_wrapper<type>)) {
         assert(CvISXSUB(cv));
+
+        if (optype == OP_METHOD) {--SP; PUTBACK; }
+
         CAIXS_accessor<type>(aTHX_ SP, cv, stash);
+
         return PL_op->op_next->op_next;
 
     } else {
-        /* we could also lift off CAIXS_entersub here, but that's a one-time action, so let it fail */
-        OP_UNSTEAL(OP_METHOD_NAMED);
+        /*
+            we could also lift off CAIXS_entersub optimization here, but that's a one-time action,
+            so let it fail on it's own
+        */
+        OP_UNSTEAL(optype);
     }
 }
 
@@ -203,6 +233,7 @@ CAIXS_install_entersub(pTHX) {
     */
 
     OP* op = PL_op;
+
     if ((op->op_spare & 1) != 1 && op->op_ppaddr == PL_ppaddr[OP_ENTERSUB] && optimize_entersub) {
         op->op_spare |= 1;
         op->op_ppaddr = &CAIXS_entersub<type>;
@@ -212,11 +243,15 @@ CAIXS_install_entersub(pTHX) {
         if (LIKELY(methop != NULL)) {   /* such op can be created by call_sv(G_METHOD_NAMED) */
             while (methop->op_sibling) { methop = methop->op_sibling; }
 
-            if (methop->op_next == op && methop->op_type == OP_METHOD_NAMED && methop->op_ppaddr == PL_ppaddr[OP_METHOD_NAMED]) {
-                methop->op_ppaddr = &CAIXS_opmethod_wrapper<type>;
+            if (methop->op_next == op) {
+                if (methop->op_type == OP_METHOD_NAMED && methop->op_ppaddr == PL_ppaddr[OP_METHOD_NAMED]) {
+                    methop->op_ppaddr = &CAIXS_opmethod_wrapper<type, OP_METHOD_NAMED>;
+                } else if (methop->op_type == OP_METHOD && methop->op_ppaddr == PL_ppaddr[OP_METHOD]) {
+                    methop->op_ppaddr = &CAIXS_opmethod_wrapper<type, OP_METHOD>;
+                }
             }
         }
-#endif
+#endif /* OPTIMIZE_OPMETHOD */
     }
 }
 
@@ -254,7 +289,7 @@ CAIXS_accessor<PrivateClass>(pTHX_ SV** SP, CV* cv, HV* stash) {
     shared_keys* keys = (shared_keys*)CAIXS_find_keys(cv);
 
     if (items > 1) {
-        SP -= items; /* no need for items == 1 case */
+        SP -= items; /* no need in the 'items == 1' case */
 
         sv_setsv(keys->storage, *(SP+2));
         PUSHs(keys->storage);
@@ -273,7 +308,7 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
     dAXMARK; dITEMS;
     SP -= items;
 
-    if (!items) croak("Usage: $obj->accessor or __PACKAGE__->accessor");
+    if (UNLIKELY(!items)) croak("Usage: $obj->accessor or __PACKAGE__->accessor");
 
     CAIXS_install_entersub<type>(aTHX);
     shared_keys* keys = CAIXS_find_keys(cv);
@@ -282,7 +317,7 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
     if (SvROK(self)) {
         HV* obj = (HV*)SvRV(self);
         if (SvTYPE((SV*)obj) != SVt_PVHV) {
-            croak("Inherited accessors can only work with object instances that is hash-based");
+            croak("Inherited accessors work only with hash-based objects");
         }
 
         if (items > 1) {
@@ -313,7 +348,7 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
 
     } else {
         GV* acc_gv = CvGV(cv);
-        if (!acc_gv) croak("Can't have pkg accessor in anon sub");
+        if (!acc_gv) croak("Can't have package accessor in anon sub");
         stash = GvSTASH(acc_gv);
 
         const char* stash_name = HvNAME(stash);
@@ -345,7 +380,7 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
             } else {
                 if (!hv_store_ent(stash, keys->pkg_key, (SV*)glob, 0)) {
                     SvREFCNT_dec_NN(glob);
-                    croak("Can't add a glob to package");
+                    croak("Couldn't add a glob to package");
                 }
             }
         }
@@ -369,8 +404,8 @@ CAIXS_accessor(pTHX_ SV** SP, CV* cv, HV* stash) {
 
     AV* supers = mro_get_linear_isa(stash);
     /*
-        First entry in 'mro_get_linear_isa' list is a 'stash' itself.
-        It's already been tested, so ajust counter and iterator to skip over it.
+        First entry in the 'mro_get_linear_isa' list is the 'stash' itself.
+        It's already been tested, so ajust both counter and iterator to skip over it.
     */
     SSize_t fill     = AvFILLp(supers);
     SV** supers_list = AvARRAY(supers);
